#!/usr/bin/env python

"""
The micropython package for processing Python source code. The code originates
from the simplify package but has had various details related to that package
removed.

Copyright (C) 2006, 2007, 2008, 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

To use this module, an importer should be constructed and the load_from_file
method used. Here, the standard path for module searching is employed:

importer = Importer(sys.path)
importer.load_from_file(filename)
importer.vacuum()

To generate programs, the above importer should be supplied in the
initialisation of a program instance, and then various methods are called:

program = Program(importer)
image = program.get_raw_image()

Such importer and program objects are the most convenient mechanism through
which the functionality of the micropython package may be accessed.
"""

from micropython.common import *
from micropython.program import Block
import micropython.ast
import micropython.data
import micropython.opt
import micropython.inspect
import micropython.table
import os

try:
    set
except NameError:
    from sets import Set as set

class Program:

    "This class supports the generation of a program image."

    supported_optimisations = micropython.opt.Optimiser.supported_optimisations

    def __init__(self, importer, optimisations=None):

        """
        Initialise the program representation with an 'importer' which is able
        to locate and load Python modules.

        The optional 'optimisations' cause certain techniques to be used in
        reducing program size and improving program efficiency.
        """

        self.importer = importer
        self.optimisations = optimisations or set()

        # Remember the tables once generated.

        self.objtable = None
        self.paramtable = None

        # Main program information.

        self.code = None
        self.code_location = None

    def get_importer(self):
        return self.importer

    # Access to finalised program information.

    def get_image(self, with_builtins=0):

        "Return a dictionary mapping modules to structures."

        if self.code is not None:
            return self.code

        objtable = self.get_object_table()
        paramtable = self.get_parameter_table()
        self.importer.finalise()

        self.code = []

        # Append constants to the image.

        for pos, const in enumerate(self.importer.constants()):
            self.code.append(const)

        last_module = self.importer.modules_ordered[-1]

        for module in self.importer.modules_ordered:
            pos = len(self.code)
            suppress_builtins = not with_builtins and module.name == "__builtins__"

            # Position the module in the image and make a translation.

            trans = micropython.ast.Translation(module, self)

            # Add header details.

            self.code.append(module)
            pos += 1

            # Append module attributes to the image.

            attributes = module.module_attributes()
            self.code += module.attributes_as_list()
            pos += len(attributes.keys())

            # Append classes and functions to the image.

            for obj in module.all_objects:
                if isinstance(obj, micropython.inspect.Class):

                    # Add header details.

                    self.code.append(obj)
                    pos += 1

                    # Append class attributes to the image.

                    attributes = obj.class_attributes()
                    self.code += obj.attributes_as_list()
                    pos += len(attributes.keys())

                    # Omit built-in function code where requested.

                    if suppress_builtins and not obj.astnode.doc is not None:
                        continue

                    # Generate the instantiator/initialiser.
                    # Append the function code to the image.

                    instantiator = obj.get_instantiator()
                    code = trans.get_instantiator_code(obj)
                    self.code += code
                    pos += len(code)

                    # Class-level code is generated separately at the module
                    # level, and the code location is set within the code
                    # generation process for the module.

                elif isinstance(obj, micropython.inspect.Function):

                    # Add header details.

                    self.code.append(obj)
                    pos += 1

                    # Append any default values to the image.
                    # Only do this for named functions (not lambdas).

                    if obj.name is not None:
                        self.code += obj.default_attrs
                        pos += len(obj.default_attrs)

                    # Omit built-in function code where requested.

                    if suppress_builtins and not obj.astnode.doc is not None:
                        pass

                    # Append the function code to the image.

                    else:
                        code = trans.get_code(obj)
                        self.code += code
                        pos += len(code)

            # Omit built-in module code where requested.

            if suppress_builtins:
                pass

            # Append the module top-level code to the image.

            else:
                code = trans.get_module_code()
                self.code += code
                pos += len(code)

        return self.code

    def get_raw_image(self, with_builtins=0):

        "Return the raw image representation of the program."

        self.get_image(with_builtins)

        objtable = self.get_object_table()
        paramtable = self.get_parameter_table()

        # Position the objects.

        pos = 0

        for item in self.code:

            # Blocks are positioned leaving space for their expansion.

            if isinstance(item, Block):
                item.location = pos
                pos += len(item.code)

            # Other multi-location objects.

            elif isinstance(item, (
                micropython.data.Class, micropython.data.Const,
                micropython.data.Function, micropython.data.Module
                )):

                if isinstance(item, micropython.data.Class):

                    # Append a template of an instance for use when
                    # instantiating classes.

                    item.instance_template_location = pos
                    pos += 1

                # Record the location of the object.

                item.location = pos
                pos += 1

                # Code and details are associated with certain objects.

                if isinstance(item, micropython.data.Function):

                    # Set the code location only where the code has been
                    # generated.

                    if not with_builtins and item.module.name == "__builtins__" and item.astnode.doc is None:
                        item.code_location = item.full_name()

                    # Skip any defaults for named functions.

                    elif item.name is not None:
                        item.code_location = pos + len(item.defaults)

                    # Skip any defaults for lambda functions.

                    else:
                        item.code_location = pos

                elif isinstance(item, micropython.data.Const):
                    pos += len(item.raw_data())

            else:
                pos += 1

        # Generate the raw code.

        self.raw_code = []

        for item in self.code:

            if isinstance(item, micropython.data.Attr):
                self.raw_code += item.as_raw(objtable, paramtable)

            elif isinstance(item, Block):
                assert item.location == len(self.raw_code)
                self.raw_code += item.as_raw(objtable, paramtable)

            elif isinstance(item, micropython.data.Class):
                assert item.instance_template_location == len(self.raw_code)
                self.raw_code += item.as_raw(objtable, paramtable,
                    with_builtins or item.module.name != "__builtins__" or item.astnode.doc is not None)
                assert item.location == len(self.raw_code) - 1

            elif isinstance(item, micropython.data.Const):
                assert item.location == len(self.raw_code)
                self.raw_code += item.as_raw(objtable, paramtable)

            elif isinstance(item, micropython.data.Function):
                assert item.location == len(self.raw_code)
                self.raw_code += item.as_raw(objtable, paramtable)

                # Check the code location only where the code has been generated.

                assert (not with_builtins and item.module.name == "__builtins__" and item.astnode.doc is None) or \
                    item.name is not None and item.code_location == len(self.raw_code) + len(item.defaults) or \
                    item.name is None and item.code_location == len(self.raw_code)

            elif isinstance(item, micropython.data.Module):
                assert item.location == len(self.raw_code)
                self.raw_code += item.as_raw(objtable, paramtable)

            else:
                self.raw_code.append(item)

            # Set the code body location for items now that the code blocks have
            # been positioned.

            if isinstance(item, (micropython.data.Class, micropython.data.Function)):
                if not with_builtins and item.module.name == "__builtins__" and item.astnode.doc is None:
                    item.code_body_location = item.full_name()
                else:
                    item.code_body_location = item.get_body_block().location

        # Fix the module locations.

        for module in self.importer.modules_ordered:

            if not with_builtins and module.name == "__builtins__":
                continue

            module.code_location = module.blocks[0].location

        self.code_location = self.importer.modules["__main__"].code_location
        return self.raw_code

    def get_object_table(self):

        "Return a table with details of attributes for classes and modules."

        if self.objtable is None:
            self.importer.vacuum()

            t = self.objtable = micropython.table.ObjectTable()
            for module in self.importer.get_modules():
                t.add(module.full_name(), module.module_attributes())

                # Add class and instance attributes for all classes, together
                # with descendant information.

                for obj in module.all_objects:
                    if isinstance(obj, micropython.inspect.Class):
                        attributes = {obj.full_name() : obj}
                        attributes.update(obj.all_attributes())
                        attributes.update(obj.all_descendants())
                        t.add(obj.full_name(), attributes)

        return self.objtable

    def get_parameter_table(self):

        "Return a table with details of parameters for functions and methods."

        # Need the object table to get at class details.

        objtable = self.get_object_table()

        if self.paramtable is None:
            t = self.paramtable = micropython.table.ParameterTable()

            # Visit each module, getting function and method details.

            for module in self.importer.get_modules():
                for obj in module.all_objects:
                    if isinstance(obj, micropython.inspect.Function):
                        t.add(obj.full_name(), obj.parameters())

                    # Classes are callable, too.
                    # Take details of the appropriate __init__ method to make an
                    # entry for an instantiation function for the class.

                    elif isinstance(obj, micropython.inspect.Class):
                        t.add(obj.get_instantiator().full_name(), obj.get_instantiator().parameters())

            # Filter out all parameter table entries not referenced by keyword
            # arguments.

            keyword_names = set()

            for module in self.importer.get_modules():
                keyword_names.update(module.keyword_names)

            for function_name, parameters in t.table.items():
                for name in parameters.keys():
                    if name in keyword_names:
                        break
                else:
                    del t.table[function_name]

        return self.paramtable

class Importer:

    "An import machine, searching for and loading modules."

    predefined_constants = {
        "None" : None,
        "True" : True,
        "False" : False,
        #"Ellipsis" : Ellipsis,
        "NotImplemented" : NotImplemented
        }

    def __init__(self, path=None, verbose=0, optimisations=None):

        """
        Initialise the importer with the given search 'path' - a list of
        directories to search for Python modules.

        The optional 'verbose' parameter causes output concerning the activities
        of the object to be produced if set to a true value (not the default).

        The optional 'optimisations' cause certain techniques to be used in
        reducing program size and improving program efficiency.
        """

        self.path = path or [os.getcwd()]
        self.verbose = verbose
        self.optimisations = optimisations or set()

        self.modules = {}
        self.modules_ordered = []
        self.loading = set()

        # Constant records.

        self.constant_values = {}
        self.constant_list = None # cache for constants
        self.init_predefined_constants()

        # Name records (used to track actual use of names).
        # Include names which may not be explicitly used in programs.
        # NOTE: Potentially declare these when inspecting.

        self.names_used = set(["__init__", "__call__", "__bool__"])

        # Status information.

        self.vacuumed = 0
        self.finalised = 0

    def get_modules(self):

        "Return all modules known to the importer."

        return self.modules.values()

    def get_module(self, name):

        "Return the module with the given 'name'."

        return self.modules[name]

    # General maintenance.

    def vacuum(self):

        "Tidy up the modules."

        if self.vacuumed:
            return

        for name, module in self.modules.items():
            if module.loaded:
                module.vacuum()
            else:
                del self.modules[name]

        self.vacuumed = 1

    def finalise(self):

        "Finalise the program."

        if self.finalised:
            return

        self.vacuum()

        for module in self.get_modules():
            module.finalise()

        self.finalised = 1

    # Name accounting.

    def use_name(self, name):

        "Register the given 'name' as being used in the program."

        self.names_used.add(name)

    # Constant accounting.

    def init_predefined_constants(self):

        "Ensure the predefined constants."

        for name, value in self.predefined_constants.items():
            self.make_constant(value)

    def get_predefined_constant(self, name):

        "Return the predefined constant for the given 'name'."

        return self.make_constant(self.predefined_constants[name])

    def get_constant(self, value):

        "Return a constant for the given 'value'."

        const = micropython.data.Const(value)
        return self.constant_values[const]

    def make_constant(self, value):

        "Make and return a constant for the given 'value'."

        # Ensure the presence of the constant's type.

        name = value.__class__.__name__
        if self.modules.has_key("__builtins__"):
            attr = self.modules["__builtins__"].get(name)
            if attr is not None:
                attr.set_referenced()

        # Make a constant object and return it.

        const = micropython.data.Const(value)
        if not self.constant_values.has_key(const):
            self.constant_values[const] = const
        return self.constant_values[const]

    def constants(self):

        "Return a list of constants."

        if self.constant_list is None:
            self.constant_list = list(self.constant_values.values())

        return self.constant_list

    # Import methods.

    def find_in_path(self, name):

        """
        Find the given module 'name' in the search path, returning None where no
        such module could be found, or a 2-tuple from the 'find' method
        otherwise.
        """

        for d in self.path:
            m = self.find(d, name)
            if m: return m
        return None

    def find(self, d, name):

        """
        In the directory 'd', find the given module 'name', where 'name' can
        either refer to a single file module or to a package. Return None if the
        'name' cannot be associated with either a file or a package directory,
        or a 2-tuple from '_find_package' or '_find_module' otherwise.
        """

        m = self._find_package(d, name)
        if m: return m
        m = self._find_module(d, name)
        if m: return m
        return None

    def _find_module(self, d, name):

        """
        In the directory 'd', find the given module 'name', returning None where
        no suitable file exists in the directory, or a 2-tuple consisting of
        None (indicating that no package directory is involved) and a filename
        indicating the location of the module.
        """

        name_py = name + os.extsep + "py"
        filename = self._find_file(d, name_py)
        if filename:
            return None, filename
        return None

    def _find_package(self, d, name):

        """
        In the directory 'd', find the given package 'name', returning None
        where no suitable package directory exists, or a 2-tuple consisting of
        a directory (indicating the location of the package directory itself)
        and a filename indicating the location of the __init__.py module which
        declares the package's top-level contents.
        """

        filename = self._find_file(d, name)
        if filename:
            init_py = "__init__" + os.path.extsep + "py"
            init_py_filename = self._find_file(filename, init_py)
            if init_py_filename:
                return filename, init_py_filename
        return None

    def _find_file(self, d, filename):

        """
        Return the filename obtained when searching the directory 'd' for the
        given 'filename', or None if no actual file exists for the filename.
        """

        filename = os.path.join(d, filename)
        if os.path.exists(filename):
            return filename
        else:
            return None

    def load(self, name, return_leaf=0):

        """
        Load the module or package with the given 'name'. Return an object
        referencing the loaded module or package, or None if no such module or
        package exists.
        """

        if self.modules.has_key(name) and self.modules[name].loaded:
            #print "Cached (%s)" % name
            return self.modules[name]
        if self.verbose:
            print "Loading", name

        # Split the name into path components, and try to find the uppermost in
        # the search path.

        path = name.split(".")
        m = self.find_in_path(path[0])
        if not m:
            if self.verbose:
                print "Not found (%s)" % path[0]
            return None # NOTE: Import error.
        d, filename = m

        # Either acquire a reference to an already-imported module, or load the
        # module from a file.

        top = module = self.load_from_file(filename, path[0])

        # For hierarchical names, traverse each path component...

        if len(path) > 1:
            if not d:
                if self.verbose:
                    print "No package (%s)" % filename
                return None # NOTE: Import error (package not found).
            else:
                self.add_submodules(d, module)

            path_so_far = path[:1]
            for p in path[1:]:
                path_so_far.append(p)

                # Find the package or module concerned.

                m = self.find(d, p)
                if not m:
                    if self.verbose:
                        print "Not found (%s)" % p
                    return None # NOTE: Import error.
                d, filename = m
                module_name = ".".join(path_so_far)

                # Either reference an imported module or load one from a file.

                submodule = self.load_from_file(filename, module_name)

                if d:
                    self.add_submodules(d, module)

                # Store the submodule within its parent module.

                module.set_module(p, submodule)
                module = submodule

        # Return either the deepest or the uppermost module.

        if return_leaf:
            return module
        else:
            return top

    def load_from_file(self, name, module_name=None):

        """
        Load the module with the given 'name' (which may be a full module path).
        """

        if module_name is None:
            module_name = "__main__"

        module = self.add_module(module_name)
        if not module.loaded and module not in self.loading:
            self.loading.add(module)
            #print "Parsing", name
            module.parse(name)
            #print "Done", name
            self.loading.remove(module)
            module.loaded = 1

        # Record the module.

        #print "Loaded", module_name, "with namespace", module.namespace.keys()
        return module

    def add_module(self, module_name):

        """
        Return the module with the given 'module_name', adding a new module
        object if one does not already exist.
        """

        if not self.modules.has_key(module_name):
            self.modules[module_name] = module = micropython.inspect.InspectedModule(module_name, self)
            self.modules_ordered.append(module)
        else:
            module = self.modules[module_name]
        return module

    def add_submodules(self, pathname, module):

        """
        Work around insufficient __all__ declarations and examine the directory
        with the given 'pathname', adding submodules to the given 'module'.
        """

        for filename in os.listdir(pathname):
            submodule, ext = os.path.splitext(filename)
            if ext not in ("", ".py"):
                continue
            module.set_module(submodule, self.add_module(module.name + "." + submodule))

# vim: tabstop=4 expandtab shiftwidth=4
